from hockey_scraper.scrape_functions import scrape_games, scrape_date_range, scrape_seasons
from hockey_scraper.live_scrape import ScrapeLiveGames, LiveGame
import hockey_scraper.scrape_functions_nwhl as nwhl



"""
Scrape the PBP info for a given game
"""
import pandas as pd
import json
import datetime
import hockey_scraper.shared as shared


def get_pbp(game_id):
    """
    Get the response for a game (e.g. https://www.nwhl.zone/game/get_play_by_plays?id=18507472)
    
    :param game_id: Given Game id (e.g. 18507472)
    
    :return: Json
    """
    page_info = {
        "url": 'https://www.nwhl.zone/game/get_play_by_plays?id={}'.format(game_id),
        "name": str(game_id),
        "type": "nwhl_json_pbp",
        "season": "nwhl",
    }
    response = shared.get_file(page_info)

    if not response:
        print("Json pbp for game {} is either not there or can't be obtained".format(game_id))
        return {}
    else:
        return json.loads(response)


def get_roster(game_json):
    """
    Parse the roster info in the game json
    
    :param game_json: Json of all the info for a game
    
    :return: Dict: Key = Player id, Value = Name
    """
    players = {}
    for player in game_json['roster_player']:
        players[player['id']] = " ".join([player['first_name'].strip(), player['last_name'].strip()])

    return players


def get_goal_players(play, event, players):
    """
    Get all the players on ice for a goal (modify play object)
    
    :param play: Our play info
    :param event: PBP play info
    :param players: Player id -> Name conversion
    
    :return: None
    """
    # Which team is plus/minus
    if play['ev_team'] == play['home_team']:
        ptype_team = {'plus': "home", 'minus': "away"}
    else:
        ptype_team = {'plus': "away", 'minus': "home"}

    # Plus and minus players
    for ptype in ['plus', 'minus']:
        for num in range(1, 7):
            play['{}Player{}_id'.format(ptype_team[ptype], num)] = event['play_actions'][1].get("{}_player_{}".format(ptype, num))

            # Control for None
            if play['{}Player{}_id'.format(ptype_team[ptype], num)]:
                play['{}Player{}'.format(ptype_team[ptype], num)] = players.get(int(play['{}Player{}_id'.format(ptype_team[ptype], num)]))
            else:
                play['{}Player{}'.format(ptype_team[ptype], num)] = None


def parse_event(event, score, teams, date, game_id, players):
    """
    Parses a single event when the info is in a json format

    :param event: json of event 
    :param score: Current score of the game
    :param teams: Teams dict (id -> name)
    :param date: date of the game
    :param game_id: game id for game
    :param players: Dict of player ids to player names
    
    :return: dictionary with the info
    """
    play = dict()

    # Basic shit
    play['play_index'] = event['play_index']
    play['date'] = date
    play['game_id'] = game_id
    play['season'] = shared.get_season(date)
    play['period'] = event['time_interval']
    play['seconds_elapsed'] = shared.convert_to_seconds(event['clock_time_string']) if event['clock_time_string'] else None
    play['home_score'], play['away_score'] = score['home'], score['away']

    # If shootout go with 'play_by_play_string' field -> more descriptive
    play['event'] = event['play_type'] if event['play_type'] != "Shootout" else event['play_by_play_string'].strip()

    # Teams
    play['home_team'], play['away_team'] = teams['home']['name'], teams['away']['name']
    if event['play_summary']['off_team_id'] == teams['home']['id']:
        play['ev_team'] = teams['home']['name']
    else:
        play['ev_team'] = teams['away']['name']

    # Player Id
    play['p1_id'] = event.get('primary_player_id')
    play['away_goalie_id'] = event['play_actions'][0].get('away_team_goalie')
    play['home_goalie_id'] = event['play_actions'][0].get('home_team_goalie')

    play['away_goalie'] = players.get(int(play['away_goalie_id']) if play['away_goalie_id'] not in ['', None] else 0)
    play['home_goalie'] = players.get(int(play['home_goalie_id']) if play['home_goalie_id'] not in ['', None] else 0)

    # Event specific stuff
    if event['play_type'] == 'Faceoff':
        play['p2_id'] = event['play_summary'].get("loser_id")
    elif event['play_type'] == 'Penalty':
        # TODO: Format better?
        play['details'] = ",".join([str(event['play_summary'].get("infraction_type", " ")),
                                    str(event['play_summary'].get("penalty_type", " ")),
                                    str(event['play_summary'].get("penalty_minutes", " "))])
    elif event['play_type'] == "Goal":
        get_goal_players(play, event, players)
        play['p2_id'] = event['play_summary'].get("assist_1_id")
        play['p3_id'] = event['play_summary'].get("assist_2_id")

        # Update Score
        if event['play_summary']['off_team_id'] == teams['home']['id']:
            score['home'] += 1
        else:
            score['away'] += 1

    # Player Id's --> Player Names
    for num in range(1, 4):
        player_id = play.get('p{num}_id'.format(num=num), 0)
        # Control for None
        player_id = player_id if player_id else 0
        play['p{num}_name'.format(num=num)] = players.get(int(player_id))

    # Coords
    play['xC'] = event['play_summary'].get('x_coord')
    play['yC'] = event['play_summary'].get('y_coord')

    return play


def parse_json(game_json, game_id,):
    """
    Scrape the json for a game
    
    plus, minus players

    :param game_json: raw json
    :param game_id: game id for game

    :return: Either a DataFrame with info for the game 
    """
    cols = ['game_id', 'date', 'season', 'period', 'seconds_elapsed', 'event', 'ev_team', 'home_team', 'away_team',
            'p1_name', 'p1_id', 'p2_name', 'p2_id', 'p3_name', 'p3_id',
            "homePlayer1", "homePlayer1_id", "homePlayer2", "homePlayer2_id", "homePlayer3", "homePlayer3_id",
            "homePlayer4", "homePlayer4_id", "homePlayer5", "homePlayer5_id", "homePlayer6", "homePlayer6_id",
            "awayPlayer1", "awayPlayer1_id", "awayPlayer2", "awayPlayer2_id", "awayPlayer3", "awayPlayer3_id",
            "awayPlayer4", "awayPlayer4_id", "awayPlayer5", "awayPlayer5_id", "awayPlayer6", "awayPlayer6_id",
            'home_goalie', 'home_goalie_id', 'away_goalie', 'away_goalie_id', 'details', 'home_score', 'away_score',
            'xC', 'yC', 'play_index']

    # B4 anything - if there are no plays we leave
    if len(game_json['plays']) == 0:
        shared.print_warning("The Json pbp for game {} contains no plays and therefore can't be parsed".format(game_id))
        return pd.DataFrame()

    # Get all the players in the game
    players = get_roster(game_json)

    # Initialize & Update as we go along
    score = {"home": 0, "away": 0}
    teams = {"home": {"id": game_json['game']['home_team'], "name": game_json['team_instance'][0]['abbrev']},
             "away": {"id": game_json['game']['away_team'], "name": game_json['team_instance'][1]['abbrev']}
             }

    # Get date from UTC timestamp
    date = game_json['plays'][0]['created_at']
    date = datetime.datetime.strptime(date[:date.rfind("-")], "%Y-%m-%dT%H:%M:%S").strftime("%Y-%m-%d")

    try:
        events = [parse_event(play, score, teams, date, game_id, players) for play in game_json['plays']]
    except Exception as e:
        shared.print_warning('Error parsing Json pbp for game {} {}'.format(game_id, e))
        return pd.DataFrame()

    df = pd.DataFrame(events, columns=cols)

    # Get rid of null events and order by play index
    df = df[(~pd.isnull(df['event'])) & (df['event'] != "")]
    df = df.sort_values(by=['play_index'])
    df = df.drop(['play_index'], axis=1)

    return df.reset_index(drop=True)


def scrape_pbp(game_id):
    """
    Scrape the pbp data for a given game
    
    :param game_id: Given Game id (e.g. 18507472)
    
    :return: DataFrame with pbp info
    """
    game_json = get_pbp(game_id)

    if not game_json:
        shared.print_warning("Json pbp for game {} is not either not there or can't be obtained".format(game_id))
        return None

    try:
        game_df = parse_json(game_json, game_id)
    except Exception as e:
        shared.print_warning('Error parsing Json pbp for game {} {}'.format(game_id, e))
        return pd.DataFrame()

    return game_df




"""
Functions to scrape by season, games, and date range
"""
import hockey_scraper.shared as shared
import hockey_scraper.json_pbp_nwhl as json_pbp
import hockey_scraper.html_schedule_nwhl as html_schedule
import pandas as pd
import random

# All columns for the pbp
cols = ['game_id', 'date', 'season', 'period', 'seconds_elapsed', 'event', 'ev_team', 'home_team', 'away_team',
        'p1_name', 'p1_id', 'p2_name', 'p2_id', 'p3_name', 'p3_id',
        "homePlayer1", "homePlayer1_id", "homePlayer2", "homePlayer2_id", "homePlayer3", "homePlayer3_id",
        "homePlayer4", "homePlayer4_id", "homePlayer5", "homePlayer5_id", "homePlayer6", "homePlayer6_id",
        "awayPlayer1", "awayPlayer1_id", "awayPlayer2", "awayPlayer2_id", "awayPlayer3", "awayPlayer3_id",
        "awayPlayer4", "awayPlayer4_id", "awayPlayer5", "awayPlayer5_id", "awayPlayer6", "awayPlayer6_id",
        'home_goalie', 'home_goalie_id', 'away_goalie', 'away_goalie_id', 'details', 'home_score', 'away_score',
        'xC', 'yC']

# Hold any games we didn't scrape for any reason
broken_games = []


def print_errors():
    """
    Print any scraping errors.
    
    :return: None
    """
    global broken_games
    if broken_games:
        print('\nBroken pbp:')
        for x in broken_games:
            print(x)

    broken_games = []


def scrape_list_of_games(games):
    """
    Scrape an arbitrary list of games given the game id's
    
    :param games: List of game_id's to scrape
    
    :return: DataFrame of pbp info 
    """
    pbp_dfs = []
    for game in games:
        print(' '.join(['Scraping NWHL Game ', str(game)]))
        pbp_df = json_pbp.scrape_pbp(game)
        if not pbp_df.empty:
            pbp_dfs.append(pbp_df)
        else:
            # Add to broken games...
            broken_games.append(game)

    # If not empty...
    if pbp_dfs:
        return pd.concat(pbp_dfs, sort=True).reset_index(drop=True)[cols]
    return None


def scrape_games(games, data_format='csv', rescrape=False, docs_dir=None):
    """
    Scrape a list of games

    :param games: list of game_ids
    :param data_format: format you want data in - csv or pandas (csv is default)
    :param rescrape: If you want to rescrape pages already scraped. Only applies if you supply a docs dir.
    :param docs_dir: Directory that either contains previously scraped docs or one that you want them to be deposited 
                     in after scraping

    :return: Dictionary with DataFrames or None
    """
    # First check if the inputs are good
    shared.check_data_format(data_format)

    # Check on the docs_dir and re_scrape
    shared.add_dir(docs_dir)
    shared.if_rescrape(rescrape)

    pbp_df = scrape_list_of_games(games)
    print_errors()

    if data_format.lower() == 'csv':
        shared.to_csv(str(random.randint(1, 101)), pbp_df, None, "nwhl")
    else:
        return pbp_df


def scrape_date_range(from_date, to_date, data_format='csv', rescrape=False, docs_dir=None):
    """
    Scrape games in given date range

    :param from_date: date you want to scrape from
    :param to_date: date you want to scrape to
    :param data_format: format you want data in - csv or pandas (csv is default)
    :param rescrape: If you want to rescrape pages already scraped. Only applies if you supply a docs dir. (def. = None)
    :param docs_dir: Directory that either contains previously scraped docs or one that you want them to be deposited 
                     in after scraping. (default is None)

    :return: Dictionary with DataFrames and errors or None
    """
    # First check if the inputs are good
    shared.check_data_format(data_format)
    shared.check_valid_dates(from_date, to_date)

    # Check on the docs_dir and re_scrape
    shared.add_dir(docs_dir)
    shared.if_rescrape(rescrape)

    # Get dates and convert to just a list of game ids
    games = html_schedule.scrape_dates(from_date, to_date)
    game_ids = [game['game_id'] for game in games]

    # Scrape all PBP
    pbp_df = scrape_list_of_games(game_ids)

    # Merge in subtype
    pbp_df = pd.merge(pbp_df, pd.DataFrame(games, columns=['game_id', 'sub_type']), on="game_id", how="left")

    print_errors()
    if data_format.lower() == 'csv':
        shared.to_csv(from_date + '--' + to_date, pbp_df, None, "nwhl")
    else:
        return pbp_df


def scrape_seasons(seasons, data_format='csv', rescrape=False, docs_dir=None):
    """
    Given list of seasons it scrapes all the seasons 

    :param seasons: list of seasons
    :param data_format: format you want data in - csv or pandas (csv is default)
    :param rescrape: If you want to rescrape pages already scraped. Only applies if you supply a docs dir.
    :param docs_dir: Directory that either contains previously scraped docs or one that you want them to be deposited 
                     in after scraping

    :return: Dictionary with DataFrames and errors or None
    """
    # First check if the inputs are good
    shared.check_data_format(data_format)

    # Check on the docs_dir and re_scrape
    shared.add_dir(docs_dir)
    shared.if_rescrape(rescrape)

    # Holds all seasons scraped (if not csv)
    master_pbps = []

    for season in seasons:
        from_date = '-'.join([str(season), '9', '1'])
        to_date = '-'.join([str(season + 1), '8', '31'])

        # Get dates and convert to just a list of game ids
        games = html_schedule.scrape_dates(from_date, to_date)
        game_ids = [game['game_id'] for game in games]

        # Scrape all PBP
        pbp_df = scrape_list_of_games(game_ids)

        # Merge in subtype
        pbp_df = pd.merge(pbp_df, pd.DataFrame(games, columns=['game_id', 'sub_type']), on="game_id", how="left")

        if data_format.lower() == 'csv':
            shared.to_csv(str(season) + str(season + 1), pbp_df, None, "nwhl")
        else:
            master_pbps.append(pbp_df)

    print_errors()
    if data_format.lower() == 'pandas':
        return pd.concat(master_pbps, sort=True)




"""
Scrape the schedule info for nwhl games
"""
import re
import datetime
from bs4 import BeautifulSoup
import hockey_scraper.shared as shared


def get_schedule(url, name):
    """
    Given a date it returns the raw html

    :param url: url for page
    :param name: Name for saved file

    :return: raw html of game
    """
    page_info = {
        "url": url,
        "name": str(name),
        "type": "html_schedule_nwhl",
        "season": "nwhl",
    }

    return shared.get_file(page_info)


def date_obj(date):
    """
    Convert date string to a datetime object
    
    :param date: e.g. 2015-10-20
    
    :return: Datetime object
    """
    return datetime.datetime.strptime(date, "%Y-%m-%d")


def get_date_games(date_info):
    """
    Get all the games corresponding to a certain date
    
    :param date_info: Page info for that date
    
    :return: All games for that day
    """
    base = "https://www.nwhl.zone/"
    response = get_schedule(base + date_info['url'], date_info['date'])
    soup = BeautifulSoup(response, "lxml")

    tables = soup.find_all("table", {"class": "statTable sortable noSortImages"})
    trs = tables[0].find("tbody").find_all("tr")

    # Each tr is a game
    games = []
    for tr in trs:
        # For finding the game id
        game_id_regex = re.compile(r"row_(\d+)")

        # Get all other game info nested inside
        tds = tr.find_all("td")
        status = tds[5].text.strip()

        # Only append if the game already started/finished
        # If scheduled will have starting time
        if status == '':
            games.append({
                "game_id": game_id_regex.findall(tr['id'])[0],
                "date": date_info['date'],
                "sub_type": date_info['sub_type'],
                "away_team": tds[0].text.strip(),
                "away_score": tds[1].text.strip(),
                "home_team": tds[2].text.strip(),
                "home_score": tds[3].text.strip(),
                "location": tds[4].text.strip(),
            })

    return games


def get_sub_dates(url, season, sub_type):
    """
    Get dates of games off of sub-season page
    
    :param url: Url for sub-season 
    :param season: Given season
    :param sub_type: e.g. Regular Season
    
    :return: List with all dates of games
    """
    soup = BeautifulSoup(get_schedule(url, "-".join([str(season), sub_type])), "lxml")
    divs = soup.find_all("div", {"class": "games-slider-group"})
    lis = [div.find_all("li") for div in divs]

    # Get all dates that have games
    dates = []
    for month in lis:
        for day in month:
            dates.append({
                "date": datetime.datetime.strptime(str(day['id'])[str(day['id']).find("day_") + len("days"):], '%Y_%m_%d').strftime('%Y-%m-%d'),
                "url": day.find("a")['href'],
                "sub_type": sub_type.strip()
            })

    return dates


def get_dates(from_date, to_date):
    """
    Get all the date pages that a game occurs in the range
    
    :param from_date: Date Scrape from
    :param to_date: Date scrape to 
    
    :return: List of Dates where games occurred 
    """
    date_range = from_date + "-" + to_date

    # Get initial info
    # Just use 2015 season
    seed_url = "https://www.nwhl.zone/schedule/day/league_instance/46947"
    soup = BeautifulSoup(get_schedule(seed_url, date_range + "-seed"), "lxml")

    # By Season (e.g. 2017-2018)
    sub_seasons = {season['label']: season.find_all("option") for season in soup.find_all("optgroup")}

    # Add Current season (here 2015 subs) - not found in above dropdown
    cur_season_subs = soup.find_all("div", {"class": "currentSeason"})[0].find_all("a")
    cur_season_subs = [sub for sub in cur_season_subs if sub['class'][0] != "close"]
    cur_season = soup.find_all("div", {"class": "currentSeason"})[0].find("span").text.strip()[:9]
    sub_seasons[cur_season] = cur_season_subs

    # Season o first date to season of last date
    # Know way to index by date so we start from the season
    from_season = shared.get_season(from_date)
    to_season = shared.get_season(to_date)

    # Get all dates for that season range (season of from_date and season of to_date)
    base = "https://www.nwhl.zone/"
    dates = []
    for season in range(from_season, to_season + 1):
        for sub in sub_seasons["-".join([str(season), str(season + 1)])]:
            # Get dates for season-sub_type combo
            # Href and value are due to current season
            try:
                sub_dates = get_sub_dates(base + sub['value'], str(season), sub.text)
            except KeyError:
                sub_dates = get_sub_dates(base + sub['href'], str(season), sub.text)
            for sub_date in sub_dates:
                # Only add dates in range
                if date_obj(sub_date['date']) >= date_obj(from_date) and date_obj(sub_date['date']) <= date_obj(to_date):
                    dates.append(sub_date)

    return dates


def scrape_dates(from_date, to_date):
    """
    Get all the games between two dates
    
    :param from_date: Date Scrape from
    :param to_date: Date scrape to
    
    :return: List of all games
    """
    games = []
    for game_date in get_dates(from_date, to_date):
        games.extend(get_date_games(game_date))

    return games

